from typing import List

ORDERED_FORCING_FIELDS: List[str] = [
    "QUANTITY",
    "FILENAME",
    "VARNAME",
    "SOURCEMASK",
    "FILETYPE",
    "METHOD",
    "EXTRAPOLATION_METHOD",
    "MAXSEARCHRADIUS",
    "OPERAND",
    "VALUE",
    "FACTOR",
    "IFRCTYP",
    "AVERAGINGTYPE",
    "RELATIVESEARCHCELLSIZE",
    "EXTRAPOLTOL",
    "PERCENTILEMINMAX",
    "AREA",
    "NUMMIN",
]
"""List[str]: List of the ordered fields names in a forcing block."""
